#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <stdint.h>

#include "physmem_lib.h"


/** True symbol for C. */
#define TRUE 1
/** Page size in words. */
#define PAGE_SIZE 1024


/** Physmem handler. */
physmemHandler_t physmemHandler;
/** Physical start address of physmem area. */
unsigned long phys=0;
/** User start address of physmem area. */
volatile unsigned long* addr=NULL;


/* PIDs of child process. */
pid_t childProcessPid=(pid_t)(-1);


/* Function for starting the child process. */
int startChildProcess(void);

/* The 'main' function of the child process. */
int childProcess(void);


/* The 'main' function of the mother process. */
int main(int argc, char* argv[])
{
  int nPages;
  unsigned long physOffset;
  unsigned long addrOffset;
  unsigned long value;

  fprintf(stderr, "[main] I am the main program. \n");
  fprintf(stderr, "[main] sizeof(unsigned long)=%d \n", (int)sizeof(unsigned long));
  fprintf(stderr, "[main] sizeof(unsigned int)=%d \n", (int)sizeof(unsigned int));

  /* Open physmem device. */
  memset(&physmemHandler, '\0', sizeof(physmemHandler_t));
  if ( physmemOpen(&physmemHandler, 1, TRUE, 0) )
  {
    fprintf(stderr, "[runRecorder] Failed to open physmem. Halting run.\n");
    return -1;
  }
  fprintf(stderr, "[main] Successfully opened physmem. \n");
  fprintf(stderr, "[main] Size of physmem: %d \n", (int)physmemHandler.phys_size_physmem);

  /* Allocate one continuous physmem page. */
  physOffset=0;
  addrOffset=0;
  phys=physmemHandler.phys_addr_physmem;
  addr=physmemHandler.addr_user_physmem;
  fprintf(stderr, "[main] Physmem base user address: %p\n", addr);
  nPages=0;
  while ( physOffset<physmemHandler.phys_size_physmem && nPages<1 )
  {
    int start_status=0;
    int end_status=0;
    if ( ( ! physmemBlockIsNotContinuous((volatile unsigned int*)(physmemHandler.addr_user_physmem+addrOffset), (volatile unsigned int*)(physmemHandler.addr_user_physmem+addrOffset+PAGE_SIZE), physmemHandler, &start_status, &end_status) ) && 
         ( ! physmemBlockCross4GB(physmemHandler.phys_size_physmem+physOffset, sizeof(unsigned int)*PAGE_SIZE) ) )
    {
      fprintf(stderr, "[main] Address of first word of first valid page: %p\n", addr+addrOffset);
//      *(addr+addrOffset)=137;
//      fprintf(stderr, "[main] Written 137 to %p\n", addr+addrOffset);
//      value=*(addr+addrOffset);
//      fprintf(stderr, "[main] Verifying value 137 under %p: %d\n", addr+addrOffset, (int)value);
      ++nPages;
    }
    physOffset+=sizeof(unsigned int)*PAGE_SIZE;
    addrOffset+=PAGE_SIZE;
  }
  if ( nPages<1 )
  {
    fprintf(stderr, "[main] Too small memory for physmem pages. Halting run.\n");
    return -1;
  }
  fprintf(stderr, "[main] Finished opening and verification of physmem.\n");

  /* Start child process. */
  fprintf(stderr, "[main] Starting child process.\n");
  if ( startChildProcess() )
  {
    fprintf(stderr, "[main] Could not start child process. Halting run.\n");
    return -1;
  }
  sleep(30);
  fprintf(stderr, "[main] Successfully finished run.\n");

  return 0;
} /* End of runRecorder */


/********************************************************
*         startChildProcess                             *
********************************************************/
int startChildProcess(void)
{
  pid_t mypid;
  /* Start child process. */
  mypid=fork();
  /* If fork failed, return with error. */
  if ( mypid<0 ) return -1;
  /* Otherwise, if we are in the childProcess side, 
     start childProcess and exit when ended. */
  else if ( mypid==(pid_t)(0) )
  {
    int childProcessExitCode=childProcess();
    exit(childProcessExitCode);
  }
  /* Otherwise, we are in the recorder process side, so 
     store PID of childProcess. */
  else childProcessPid=mypid;
  return 0;
} /* End of startChildProcess */


/********************************************************
*         childProcess                                  *
********************************************************/
int childProcess(void)
{
  unsigned long physOffset;
  unsigned long addrOffset;
  int nPages;
  unsigned long value=0;

  /* Touch first word in the first continuous physmem page. */
  fprintf(stderr, "[childProcess] I am the child process.\n");
  fprintf(stderr, "[childProcess] Physmem base user address: %p\n", addr);
  physOffset=0;
  addrOffset=0;
  nPages=0;
  while ( physOffset<physmemHandler.phys_size_physmem && nPages<1 )
  {
    int start_status=0;
    int end_status=0;
    if ( ( ! physmemBlockIsNotContinuous((volatile unsigned int*)(addr+addrOffset), (volatile unsigned int*)(addr+addrOffset+PAGE_SIZE), physmemHandler, &start_status, &end_status) ) && 
         ( ! physmemBlockCross4GB(phys+physOffset, sizeof(unsigned int)*PAGE_SIZE) ) )
    {
      fprintf(stderr, "[childProcess] Address of first word of first valid page: %p\n", addr+addrOffset);
      fprintf(stderr, "[childProcess] Verifying value under %p\n", addr+addrOffset);
      fprintf(stderr, "               (if not 137 is printed below, it means I died with segfault):\n");
      value=*(addr+addrOffset); /* <- Here I segfault. */
      fprintf(stderr, "               %d \n", (int)value);
      ++nPages;
    }
    physOffset+=sizeof(unsigned int)*PAGE_SIZE;
    addrOffset+=PAGE_SIZE;
  }
  if ( nPages<1 )
  {
    fprintf(stderr, "[childProcess] Too small memory for physmem pages. Halting run.\n");
    return -1;
  }
  sleep(10);
  fprintf(stderr, "[childProcess] Successfully finished run.\n");

  return 0;

} /* End of childProcess */



